@@ -56,8 +56,7 @@ feature_presets = {}
 
 # Full stack
 feature_presets['neurosymbolic'] = [
-    'concept_features', 'document_graph', 'neural_features_dbow',
-    'neural_features_dm', 'relational_features_token', 'topic_features',
+    'concept_features', 'document_graph', 'relational_features_token', 'topic_features',
     'keyword_features', 'relational_features_char', 'char_features',
     'word_features', 'relational_features_bigram', 'contextual_features'
 ]
