@@ -17,11 +17,12 @@ def run():
         reptype,  ## See the documentation for all possible representation types.
         n_fold_cv=3,
         memory_storage="memory",
+        verbose=0,
         sparsity=0.1,
         visualize_progress=True, ## Stores progress as PROGRESS_{generation}.pdf file
         upsample=
         False,  ## Suitable for imbalanced data - randomized upsampling tends to help.
-        time_constraint=0.1).evolve(
+        time_constraint=0.2).evolve(
             strategy="evolution"
         )  ## strategy = "direct-learning" trains a single learner.
 
@@ -45,7 +46,7 @@ def run():
     output_classification_results(
         predictions,
         test_classes,
-        f"./predictions/test_{jid}_autobot{reptype}_report.json",
+        f"./predictions/SKLEARN.json",
         model_spec={})
 
 
